@@ -1,16 +1,18 @@
-/*
+/*! \file
 
-Copyright 2011 University Corporation for Atmospheric
+Copyright 2018 University Corporation for Atmospheric
 Research/Unidata. See \ref copyright file for more info.  */
 
-#include <config.h>
+#include "config.h"
 #include <stdio.h>
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #endif
-#ifdef _MSC_VER /* Microsoft Compilers */
-#include <io.h>
+#ifdef _MSC_VER
+#include "XGetopt.h"
+#define snprintf _snprintf
 #endif
+
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -18,23 +20,16 @@ Research/Unidata. See \ref copyright file for more info.  */
 #include <fcntl.h>
 #endif
 
-#ifdef _MSC_VER
-#define snprintf _snprintf
-#include "XGetopt.h"
-int opterr;
-int optind;
-#endif
-
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
 #include <math.h>
-#ifdef HAVE_LOCALE_H
-#include <locale.h>
-#endif	/* HAVE_LOCALE_H */
+
 #include "netcdf.h"
 #include "netcdf_mem.h"
+#include "netcdf_filter.h"
+#include "netcdf_aux.h"
 #include "utils.h"
 #include "nccomps.h"
 #include "nctime0.h"		/* new iso time and calendar stuff */
@@ -44,14 +39,18 @@ int optind;
 #include "indent.h"
 #include "isnan.h"
 #include "cdl.h"
+#include "nclog.h"
+#include "ncpathmgr.h"
+#include "nclist.h"
+#include "ncuri.h"
+#include "nc_provenance.h"
+#include "ncpathmgr.h"
 
 #ifdef USE_NETCDF4
 #include "nc4internal.h" /* to get name of the special properties file */
 #endif
 
-#if !defined(HAVE_SSIZE_T) && !defined(H5_SIZEOF_SSIZE_T)
-typedef int ssize_t;
-#endif
+extern int nc__testurl(const char*,char**);
 
 #define XML_VERSION "1.0"
 
@@ -62,8 +61,8 @@ typedef int ssize_t;
    we need to be careful about printing their attributes.
 */
 static const char* keywords[] = {
-"variable",
-"dimension",
+"variables",
+"dimensions",
 "data",
 "group",
 "types",
@@ -144,45 +143,35 @@ usage(void)
 static char *
 name_path(const char *path)
 {
-    const char *cp;
-    char *new;
-    char *sp;
+    char* cvtpath = NULL;
+    const char *cp = NULL;
+    char *sp = NULL;
+    size_t cplen = 0;
+    char* base = NULL;
 
-#ifdef vms
-#define FILE_DELIMITER ']'
-#endif
-#if defined(WIN32) || defined(msdos)
-#define FILE_DELIMITER '\\'
-#endif
-#ifndef FILE_DELIMITER /* default to unix */
-#define FILE_DELIMITER '/'
-#endif
+    if((cvtpath = NCpathcvt(path))==NULL)
+        return NULL;
 
-#ifdef USE_DAP
     /* See if this is a url */
-    {
-	char* base;
+    if(nc__testurl(cvtpath,&base))
+ 	 goto done; /* Looks like a url */
+    /* else fall thru and treat like a file path */
 
-        extern int nc__testurl(const char*,char**);
-
-
- 	if(nc__testurl(path,&base)) {
- 	    return base; /* Looks like a url */
-	}
-	/* else fall thru and treat like a file path */
-    }
-#endif /*USE_DAP*/
-
-    cp = strrchr(path, FILE_DELIMITER);
-    if (cp == 0)		/* no delimiter */
-      cp = path;
+    cp = strrchr(cvtpath, '/');
+    if (cp == NULL)		/* no delimiter */
+      cp = cvtpath;
     else			/* skip delimiter */
       cp++;
-    new = (char *) emalloc((unsigned) (strlen(cp)+1));
-    (void) strncpy(new, cp, strlen(cp) + 1);	/* copy last component of path */
-    if ((sp = strrchr(new, '.')) != NULL)
+    cplen = strlen(cp);
+    base = (char *) emalloc((unsigned) (cplen+1));
+    base[0] = '\0';
+    strlcat(base,cp,cplen+1);
+    if ((sp = strrchr(base, '.')) != NULL)
       *sp = '\0';		/* strip off any extension */
-    return new;
+
+done:
+    nullfree(cvtpath);
+    return base;
 }
 
 /* Return primitive type name */
@@ -273,7 +262,7 @@ kind_string(int kind)
     case NC_FORMAT_NETCDF4_CLASSIC:
 	return "netCDF-4 classic model";
     default:
-	error("unrecognized file format: %d");
+       error("unrecognized file format: %d", kind);
 	return "unrecognized";
     }
 }
@@ -319,7 +308,7 @@ kind_string_extended(int kind, int mode)
     return text;
 }
 
-#ifdef USE_DISKLESS
+#if 0
 static int
 fileopen(const char* path, void** memp, size_t* sizep)
 {
@@ -338,9 +327,9 @@ fileopen(const char* path, void** memp, size_t* sizep)
 #endif
     oflags |= O_EXCL;
 #ifdef vms
-    fd = open(path, oflags, 0, "ctx=stm");
+    fd = NCopen3(path, oflags, 0, "ctx=stm");
 #else
-    fd  = open(path, oflags);
+    fd  = NCopen2(path, oflags);
 #endif
     if(fd < 0) {
 	status = errno;
@@ -763,11 +752,11 @@ pr_att(
     ncatt_t att;			/* attribute */
 
     NC_CHECK( nc_inq_attname(ncid, varid, ia, att.name) );
-#ifdef ENABLE_FILEINFO
+#ifdef USE_NETCDF4
     if (ncid == getrootid(ncid)
         && varid == NC_GLOBAL
         && strcmp(att.name,NCPROPS)==0)
-	return; /* will be printed elsewere */
+	return; /* will be printed elsewhere */
 #endif
     NC_CHECK( nc_inq_att(ncid, varid, att.name, &att.type, &att.len) );
     att.tinfo = get_typeinfo(att.type);
@@ -835,7 +824,7 @@ pr_att(
        size_t type_size, nfields;
        nc_type base_nc_type;
        int class, i;
-       void *data;
+       void *data = NULL;
 
        NC_CHECK( nc_inq_user_type(ncid, att.type,  type_name, &type_size,
 				  &base_nc_type, &nfields, &class));
@@ -844,7 +833,7 @@ pr_att(
 	  case NC_VLEN:
 	      /* because size returned for vlen is base type size, but we
 	       * need space to read array of vlen structs into ... */
-	      data = emalloc((att.len + 1) * sizeof(nc_vlen_t));
+              data = emalloc((att.len + 1) * sizeof(nc_vlen_t));
 	     break;
 	  case NC_OPAQUE:
 	      data = emalloc((att.len + 1) * type_size);
@@ -865,7 +854,6 @@ pr_att(
        switch(class) {
        case NC_VLEN:
 	   pr_any_att_vals(&att, data);
-	   free(data);
 	   break;
        case NC_OPAQUE: {
 	   char *sout = emalloc(2 * type_size + strlen("0X") + 1);
@@ -876,8 +864,7 @@ pr_att(
 	       cp += type_size;
 	   }
 	   free(sout);
-       }
-	   break;
+         } break;
        case NC_ENUM: {
 	   int64_t value;
 	   for (i = 0; i < att.len; i++) {
@@ -917,15 +904,15 @@ pr_att(
 	       print_name(enum_name);
 	       printf("%s", i < att.len-1 ? ", " : "");
 	   }
-       }
-	   break;
+         } break;
        case NC_COMPOUND:
 	   pr_any_att_vals(&att, data);
-	   free(data);
 	   break;
        default:
 	   error("unrecognized class of user defined type: %d", class);
        }
+       ncaux_reclaim_data(ncid,att.type,data,att.len);
+       free(data);
        printf (" ;");		/* terminator for user defined types */
     }
 #endif /* USE_NETCDF4 */
@@ -980,17 +967,19 @@ pr_att_specials(
     const ncvar_t *varp
     )
 {
+    int contig = NC_CHUNKED;
     /* No special variable attributes for classic or 64-bit offset data */
     if(kind == 1 || kind == 2)
 	return;
-    /* _Chunking */
-    if (varp->ndims > 0) {	/* no chunking for scalar variables */
-	int contig = 0;
-	NC_CHECK( nc_inq_var_chunking(ncid, varid, &contig, NULL ) );
-	if(contig == 1) {
-	    pr_att_name(ncid, varp->name, NC_ATT_STORAGE);
+    /* _Chunking tests */
+    NC_CHECK( nc_inq_var_chunking(ncid, varid, &contig, NULL ) );
+    if(contig == NC_CONTIGUOUS) {
+  	    pr_att_name(ncid, varp->name, NC_ATT_STORAGE);
 	    printf(" = \"contiguous\" ;\n");
-	} else {
+    } else if(contig == NC_COMPACT) {
+	    pr_att_name(ncid, varp->name, NC_ATT_STORAGE);
+	    printf(" = \"compact\" ;\n");
+    } else {
  	   size_t *chunkp;
 	   int i;
 	    pr_att_name(ncid, varp->name, NC_ATT_STORAGE);
@@ -1004,18 +993,64 @@ pr_att_specials(
 		printf("%lu%s", (unsigned long)chunkp[i], i+1 < varp->ndims ? ", " : " ;\n");
 	    }
 	    free(chunkp);
-	}
     }
 
-    /*_Deflate, _Shuffle */
+    /* _Filter (including deflate and shuffle) */
     {
-	int shuffle=NC_NOSHUFFLE, deflate=0, deflate_level=0;
-	NC_CHECK( nc_inq_var_deflate(ncid, varid, &shuffle,
-				     &deflate, &deflate_level) );
-	if(deflate != 0) {
-	    pr_att_name(ncid, varp->name, NC_ATT_DEFLATE);
-	    printf(" = %d ;\n", deflate_level);
+	size_t nparams, nfilters, nbytes;
+	int shuffle=NC_NOSHUFFLE;
+	unsigned int* filterids = NULL;
+	unsigned int* params = NULL;
+	int usedeflateatt = 0;
+
+	/* Get applicable filter ids */
+	NC_CHECK(nc_inq_var_filter_ids(ncid, varid, &nfilters, NULL));
+	/* Get set of filters for this variable */
+	if(nfilters > 0) {
+	    filterids = (unsigned int*)malloc(sizeof(unsigned int)*nfilters);
+	    if(filterids == NULL) NC_CHECK(NC_ENOMEM);
+	} else
+	    filterids = NULL;
+	NC_CHECK(nc_inq_var_filter_ids(ncid, varid, &nfilters, filterids));
+        if(nfilters > 0) {
+	    int k;
+	    int pratt = 0;
+	    for(k=0;k<nfilters;k++) {
+		NC_CHECK(nc_inq_var_filter_info(ncid, varid, filterids[k], &nparams, NULL));
+	        if(nparams > 0) {
+  	            params = (unsigned int*)calloc(1,sizeof(unsigned int)*nparams);
+	            NC_CHECK(nc_inq_var_filter_info(ncid, varid, filterids[k], &nbytes, params));
+		} else
+		    params = NULL;
+	        /* Use _Deflate if the first filter is zip */
+		if(k == 0 && filterids[k] == H5Z_FILTER_DEFLATE) {
+	            pr_att_name(ncid, varp->name, NC_ATT_DEFLATE);
+	            printf(" = %d", (int)params[0]);
+		    pratt = 1;
+		    usedeflateatt = 1;
+       	            nullfree(params); params = NULL;
+		    continue;
+		}
+		if(pratt || k == 0) {
+		    pr_att_name(ncid,varp->name,NC_ATT_FILTER);
+		    printf(" = \"");
+		    pratt = 0;
+		}
+	        if(k > 0) printf("|");
+		printf("%u",filterids[k]);
+		if(nparams > 0) {
+	            int i;
+		    for(i=0;i<nparams;i++)
+		        printf(",%u",params[i]);
+	        }
+       	        nullfree(params); params = NULL;
+	    }
+            if(!usedeflateatt) printf("\"");
+            printf(" ;\n");
 	}
+	if(filterids) free(filterids);
+        /* Finally, do Shuffle */
+	NC_CHECK( nc_inq_var_deflate(ncid, varid, &shuffle, NULL, NULL));
 	if(shuffle != NC_NOSHUFFLE) {
 	    pr_att_name(ncid, varp->name, NC_ATT_SHUFFLE);
 	    printf(" = \"true\" ;\n");
@@ -1074,7 +1109,7 @@ pr_att_specials(
 }
 #endif /* USE_NETCDF4 */
 
-#ifdef ENABLE_FILEINFO /*=>NETCDF4*/
+
 static void
 pr_att_hidden(
     int ncid,
@@ -1083,22 +1118,26 @@ pr_att_hidden(
 {
     int stat;
     size_t len;
-    char propdata[NCPROPS_LENGTH];
 
     /* No special variable attributes for classic or 64-bit offset data */
-    if(kind == 1 || kind == 2)
-	return;
+#if 0
+    if(kind == 1 || kind == 2) return;
+#endif
     /* Print out Selected hidden attributes */
     /* NCPROPS */
     stat = nc_inq_att(ncid,NC_GLOBAL,NCPROPS,NULL,&len);
-    if(stat == NC_NOERR && len < sizeof(propdata)) {
+    if(stat == NC_NOERR) {
+	char* propdata = (char*)malloc(len+1);
+	if(propdata == NULL)
+	    return;
         stat = nc_get_att_text(ncid,NC_GLOBAL,NCPROPS,propdata);
         if(stat == NC_NOERR) {
             pr_att_name(ncid, "", NCPROPS);
             /* make sure its null terminated */
-            propdata[NCPROPS_LENGTH-1] = '\0';
+            propdata[len] = '\0';
             printf(" = \"%s\" ;\n",propdata);
         }
+	free(propdata);
     }
     /* _SuperblockVersion */
     stat = nc_inq_att(ncid,NC_GLOBAL,SUPERBLOCKATT,NULL,&len);
@@ -1121,7 +1160,6 @@ pr_att_hidden(
         }
     }
 }
-#endif /* ENABLE_FILEINFO */
 
 /*
  * Print a variable attribute for NcML
@@ -1138,7 +1176,7 @@ pr_attx(
     int attvalslen = 0;
 
     NC_CHECK( nc_inq_attname(ncid, varid, ia, att.name) );
-#ifdef ENABLE_FILEINFO
+#ifdef USE_NETCDF4
     if (ncid == getrootid(ncid)
 	&& varid == NC_GLOBAL
         && strcmp(att.name,NCPROPS)==0
@@ -1258,6 +1296,7 @@ print_enum_type(int ncid, nc_type typeid) {
     char safe_buf[SAFE_BUF_LEN];
     char *delim;
     int64_t data;	    /* space for data of any primitive type */
+    void* raw;
     char *esc_btn;
     char *esc_tn;
     char *esc_mn;
@@ -1280,30 +1319,31 @@ print_enum_type(int ncid, nc_type typeid) {
 	if (f == type_nfields - 1)
 	    delim = "} ;\n";
 	NC_CHECK( nc_inq_enum_member(ncid, typeid, f, memname, &data) );
+	raw = (void*)&data;
 	switch (base_nc_type) {
 	case NC_BYTE:
-	    memval = *(char *)&data;
+	    memval = *(char *)raw;
 	    break;
 	case NC_SHORT:
-	    memval = *(short *)&data;
+	    memval = *(short *)raw;
 	    break;
 	case NC_INT:
-	    memval = *(int *)&data;
+	    memval = *(int *)raw;
 	    break;
 	case NC_UBYTE:
-	    memval = *(unsigned char *)&data;
+	    memval = *(unsigned char *)raw;
 	    break;
 	case NC_USHORT:
-	    memval = *(unsigned short *)&data;
+	    memval = *(unsigned short *)raw;
 	    break;
 	case NC_UINT:
-	    memval = *(unsigned int *)&data;
+	    memval = *(unsigned int *)raw;
 	    break;
 	case NC_INT64:
-	    memval = *(int64_t *)&data;
+	    memval = *(int64_t *)raw;
 	    break;
 	case NC_UINT64:
-	    memval = *(uint64_t *)&data;
+	    memval = *(uint64_t *)raw;
 	    break;
 	default:
 	    error("Bad base type for enum!");
@@ -1410,7 +1450,8 @@ print_ud_type(int ncid, nc_type typeid) {
 #endif /* USE_NETCDF4 */
 
 static void
-get_fill_info(int ncid, int varid, ncvar_t *vp) {
+get_fill_info(int ncid, int varid, ncvar_t *vp)
+{
     ncatt_t att;			/* attribute */
     int nc_status;			/* return from netcdf calls */
     void *fillvalp = NULL;
@@ -1419,7 +1460,7 @@ get_fill_info(int ncid, int varid, ncvar_t *vp) {
 
     /* get _FillValue attribute */
     nc_status = nc_inq_att(ncid,varid,_FillValue,&att.type,&att.len);
-    fillvalp = emalloc(vp->tinfo->size + 1);
+    fillvalp = ecalloc(vp->tinfo->size + 1);
     if(nc_status == NC_NOERR &&
        att.type == vp->type && att.len == 1) {
 	NC_CHECK(nc_get_att(ncid, varid, _FillValue, fillvalp));
@@ -1465,9 +1506,20 @@ get_fill_info(int ncid, int varid, ncvar_t *vp) {
 	    *(uint64_t *)fillvalp = NC_FILL_UINT64;
 	    break;
 #ifdef USE_NETCDF4
-	case NC_STRING:
-	    *((char **)fillvalp) = strdup(NC_FILL_STRING);
-	    break;
+	case NC_STRING: {
+	    char* s;
+	    size_t len = strlen(NC_FILL_STRING);
+#if 0
+	    /* In order to avoid mem leak, allocate this string as part of fillvalp */
+            fillvalp = erealloc(fillvalp, vp->tinfo->size + 1 + len + 1);
+	    s = ((char*)fillvalp) + vp->tinfo->size + 1;
+#else
+	    s = malloc(len+1);
+#endif
+	    memcpy(s,NC_FILL_STRING,len);
+	    s[len] = '\0';
+	    *((char **)fillvalp) = s;
+	    } break;
 #endif /* USE_NETCDF4 */
 	default:		/* no default fill values for NC_NAT
 				   or user-defined types */
@@ -1480,7 +1532,6 @@ get_fill_info(int ncid, int varid, ncvar_t *vp) {
     vp->fillvalp = fillvalp;
 }
 
-
 /* Recursively dump the contents of a group. (Only netcdf-4 format
  * files can have groups, so recursion will not take place for classic
  * format files.)
@@ -1657,10 +1708,10 @@ do_ncdump_rec(int ncid, const char *path)
     * to know how to print strings with embedded newlines. */
    NC_CHECK( nc_inq_format(ncid, &kind) );
 
-   /* For each var, get and print out info. */
-
    memset((void*)&var,0,sizeof(var));
 
+   /* For each var, get and print out info. */
+
    for (varid = 0; varid < nvars; varid++) {
       NC_CHECK( nc_inq_varndims(ncid, varid, &var.ndims) );
       if(var.dims != NULL) free(var.dims);
@@ -1697,17 +1748,27 @@ do_ncdump_rec(int ncid, const char *path)
 	  * ambiguous. */
 	 {
 	     int dimid_test;	/* to see if dimname is ambiguous */
+	     int target_dimid;  /* from variable dim list */
 	     int locid;		/* group id where dimension is defined */
+	     /*Locate the innermost definition of a dimension with given name*/
 	     NC_CHECK( nc_inq_dimid(ncid, dim_name, &dimid_test) );
+
+	     /* Now, starting with current group, walk the parent chain
+                upward looking for the target dim_id */
+	     target_dimid = var.dims[id];
 	     locid = ncid;
-	     while(var.dims[id] != dimid_test) { /* not in locid, try ancestors */
+	     while(target_dimid != dimid_test) {/*not in locid, try ancestors*/
 		 int parent_id;
 		 NC_CHECK( nc_inq_grp_parent(locid, &parent_id) );
 		 locid = parent_id;
+		 /* Is dim of this name defined in this group or higher? */
 		 NC_CHECK( nc_inq_dimid(locid, dim_name, &dimid_test) );
 	     }
-	     /* dimid is in group locid, prefix dimname with group name if needed */
-	     if(locid != ncid) {
+	     /* innermost dimid with given name is in group locid.
+                If this is not current group, then use fully qualified
+                name (fqn) for the dimension name by prefixing dimname
+                with group name */
+	     if(locid != ncid) { /* We need to use fqn */
 		 size_t len;
 		 char *locname;	/* the group name */
 		 NC_CHECK( nc_inq_grpname_full(locid, &len, NULL) );
@@ -1751,9 +1812,8 @@ do_ncdump_rec(int ncid, const char *path)
    }
    if (is_root && formatting_specs.special_atts) { /* output special attribute
 					   * for format variant */
-#ifdef ENABLE_FILEINFO
-       pr_att_hidden(ncid, kind);
-#endif
+
+     pr_att_hidden(ncid, kind);
        pr_att_global_format(ncid, kind);
    }
 
@@ -1806,7 +1866,6 @@ do_ncdump_rec(int ncid, const char *path)
 	     vdims = 0;
 	     continue;
 	 }
-	 if(var.fillvalp != NULL) free(var.fillvalp);
 	 get_fill_info(ncid, varid, &var); /* sets has_fillval, fillvalp mmbrs */
 	 if(var.timeinfo != NULL) {
 	     if(var.timeinfo->units) free(var.timeinfo->units);
@@ -1821,6 +1880,8 @@ do_ncdump_rec(int ncid, const char *path)
 	    error("can't output data for variable %s", var.name);
 	    goto done;
 	 }
+	 if(var.fillvalp != NULL)
+	     {ncaux_reclaim_data(ncid,var.tinfo->tid,var.fillvalp,1); free(var.fillvalp); var.fillvalp = NULL;}
       }
       if (vdims) {
 	  free(vdims);
@@ -1871,7 +1932,12 @@ do_ncdump_rec(int ncid, const char *path)
 
 done:
    if(var.dims != NULL) free(var.dims);
-   if(var.fillvalp != NULL) free(var.fillvalp);
+   if(var.fillvalp != NULL) {
+	/* Release any data hanging off of fillvalp */
+	ncaux_reclaim_data(ncid,var.tinfo->tid,var.fillvalp,1);
+	free(var.fillvalp);
+	var.fillvalp = NULL;
+   }
    if(var.timeinfo != NULL) {
       if(var.timeinfo->units) free(var.timeinfo->units);
       free(var.timeinfo);
@@ -1997,6 +2063,8 @@ do_ncdumpx(int ncid, const char *path)
 	freeidlist(vlist);
     if(dims)
 	free(dims);
+    if(var.dims != NULL)
+        free(var.dims);
 }
 
 /*
@@ -2067,20 +2135,22 @@ set_precision(const char *optarg)
 
 
 #ifdef USE_DAP
-#define DAP_CLIENT_CACHE_DIRECTIVE	"[cache]"
+#define DAP_CLIENT_CACHE_DIRECTIVE	"cache"
 /* replace path string with same string prefixed by
  * DAP_CLIENT_NCDUMP_DIRECTIVE */
-static
-void adapt_url_for_cache(char **pathp) {
-    char prefix[] = DAP_CLIENT_CACHE_DIRECTIVE;
+static void
+adapt_url_for_cache(char **pathp)
+{
     char* path = *pathp;
-    char *tmp_path = strdup(path);
-    path = (char *)emalloc(strlen(prefix) + strlen(tmp_path) + 1);
-    path[0] = '\0';
-    strncat(path, prefix, strlen(prefix));
-    strncat(path, tmp_path, strlen(tmp_path));
-    free(tmp_path);
-    *pathp = path;
+    NCURI* url = NULL;
+    ncuriparse(path,&url);
+    if(url == NULL) return;
+    ncuriappendfragmentkey(url,DAP_CLIENT_CACHE_DIRECTIVE,NULL);
+    if(*pathp) free(*pathp);
+    path = ncuribuild(url,NULL,NULL,NCURIALL);
+    if(pathp) {*pathp = path; path = NULL;}
+    ncurifree(url);
+    nullfree(path);
     return;
 }
 #endif
@@ -2088,6 +2158,7 @@ void adapt_url_for_cache(char **pathp) {
 int
 main(int argc, char *argv[])
 {
+    int ncstat = NC_NOERR;
     int c;
     int i;
     int max_len = 80;		/* default maximum line length */
@@ -2096,15 +2167,15 @@ main(int argc, char *argv[])
     bool_t kind_out = false;	/* if true, just output kind of netCDF file */
     bool_t kind_out_extended = false;	/* output inq_format vs inq_format_extended */
     int Xp_flag = 0;    /* indicate that -Xp flag was set */
+    char* path = NULL;
+    char errmsg[4096];
 
-#if defined(WIN32) || defined(msdos) || defined(WIN64)
+    errmsg[0] = '\0';
+
+#if defined(_WIN32) || defined(msdos) || defined(WIN64)
     putenv("PRINTF_EXPONENT_DIGITS=2"); /* Enforce unix/linux style exponent formatting. */
 #endif
 
-#ifdef HAVE_LOCALE_H
-    setlocale(LC_ALL, "C");     /* CDL may be ambiguous with other locales */
-#endif /* HAVE_LOCALE_H */
-    opterr = 1;
     progname = argv[0];
     set_formats(FLT_DIGITS, DBL_DIGITS); /* default for float, double data */
 
@@ -2116,6 +2187,7 @@ main(int argc, char *argv[])
        exit(EXIT_SUCCESS);
     }
 
+    opterr = 1;
     while ((c = getopt(argc, argv, "b:cd:f:g:hikl:n:p:stv:xwKL:X:")) != EOF)
       switch(c) {
 	case 'h':		/* dump header only, no data */
@@ -2141,7 +2213,8 @@ main(int argc, char *argv[])
 	      formatting_specs.data_lang = LANG_F;
 	      break;
 	    default:
-	      error("invalid value for -b option: %s", optarg);
+	      snprintf(errmsg,sizeof(errmsg),"invalid value for -b option: %s", optarg);
+	      goto fail;
 	  }
 	  break;
 	case 'f':		/* full comments in data section */
@@ -2154,13 +2227,15 @@ main(int argc, char *argv[])
 	      formatting_specs.data_lang = LANG_F;
 	      break;
 	    default:
-	      error("invalid value for -f option: %s", optarg);
+	      snprintf(errmsg,sizeof(errmsg),"invalid value for -f option: %s", optarg);
+	      goto fail;
 	  }
 	  break;
 	case 'l':		/* maximum line length */
 	  max_len = (int) strtol(optarg, 0, 0);
 	  if (max_len < 10) {
-	      error("unreasonably small line length specified: %d", max_len);
+	      snprintf(errmsg,sizeof(errmsg),"unreasonably small line length specified: %d", max_len);
+	      goto fail;
 	  }
 	  break;
 	case 'v':		/* variable names */
@@ -2213,8 +2288,8 @@ main(int argc, char *argv[])
 	      Xp_flag = 1; /* record that this flag was set */
 	      break;
 	    default:
-	      error("invalid value for -X option: %s", optarg);
-	      break;
+	      snprintf(errmsg,sizeof(errmsg),"invalid value for -X option: %s", optarg);
+	      goto fail;
 	  }
 	  break;
         case 'L':
@@ -2225,6 +2300,7 @@ main(int argc, char *argv[])
 	    nc_set_log_level(level);
 	  }
 #endif
+	  ncsetlogging(1);
 	  break;
         case '?':
 	  usage();
@@ -2257,40 +2333,42 @@ main(int argc, char *argv[])
 
     init_epsilons();
 
-    {
-	char *path = strdup(argv[i]);
-	if(!path)
-	    error("out of memory copying argument %s", argv[i]);
-        if (!nameopt)
-	    formatting_specs.name = name_path(path);
-	if (argc > 0) {
-	    int ncid, nc_status;
-	    /* If path is a URL, prefix with client-side directive to
-	     * make ncdump reasonably efficient */
+    /* We need to look for escape characters because the argument
+       may have come in via a shell script */
+    path = NC_backslashUnescape(argv[i]);
+    if(path == NULL) {
+	snprintf(errmsg,sizeof(errmsg),"out of memory un-escaping argument %s", argv[i]);
+	goto fail;
+    }
+
+    if (!nameopt)
+        formatting_specs.name = name_path(path);
+    if (argc > 0) {
+        int ncid;
+	/* If path is a URL, do some fixups */
+	if(nc__testurl(path, NULL)) {/* See if this is a url */
+	    /*  Prefix with client-side directive to
+             * make ncdump reasonably efficient */
 #ifdef USE_DAP
-	    if(formatting_specs.with_cache) /* by default, don't use cache directive */
-	    {
-		extern int nc__testurl(const char*,char**);
-		/* See if this is a url */
-		if(nc__testurl(path, NULL)) {
-		    adapt_url_for_cache(&path);
-		}
-		/* else fall thru and treat like a file path */
+	    if(formatting_specs.with_cache) { /* by default, don't use cache directive */
+	        adapt_url_for_cache(&path);
 	    }
-#endif /*USE_DAP*/
-#ifdef USE_DISKLESS
-	    if(formatting_specs.xopt_inmemory) {
+#endif
+	} /* else fall thru and treat like a file path */
+        if(formatting_specs.xopt_inmemory) {
+#if 0
 		size_t size = 0;
 		void* mem = NULL;
-		nc_status = fileopen(path,&mem,&size);
-		if(nc_status == NC_NOERR)
-	            nc_status = nc_open_mem(path,NC_DISKLESS|NC_INMEMORY,size,mem,&ncid);
-	    } else
+		ncstat = fileopen(path,&mem,&size);
+		if(ncstat == NC_NOERR)
+	            ncstat = nc_open_mem(path,NC_INMEMORY,size,mem,&ncid);
+#else
+	        ncstat = nc_open(path,NC_DISKLESS|NC_NOWRITE,&ncid);
 #endif
-	        nc_status = nc_open(path, NC_NOWRITE, &ncid);
-	    if (nc_status != NC_NOERR) {
-		error("%s: %s", path, nc_strerror(nc_status));
-	    }
+	    } else /* just a file */
+	        ncstat = nc_open(path, NC_NOWRITE, &ncid);
+	    if (ncstat != NC_NOERR) goto fail;
+
 	    NC_CHECK( nc_inq_format(ncid, &formatting_specs.nc_kind) );
 	    NC_CHECK( nc_inq_format_extended(ncid,
                                              &formatting_specs.nc_extended,
@@ -2303,21 +2381,21 @@ main(int argc, char *argv[])
 		/* Initialize list of types. */
 		init_types(ncid);
 		/* Check if any vars in -v don't exist */
-		if(missing_vars(ncid, formatting_specs.nlvars, formatting_specs.lvars))
-		    exit(EXIT_FAILURE);
+		if(missing_vars(ncid, formatting_specs.nlvars, formatting_specs.lvars)) {
+		    snprintf(errmsg,sizeof(errmsg),"-v: non-existent variables");
+		    goto fail;
+		}
 		if(formatting_specs.nlgrps > 0) {
-		    if(formatting_specs.nc_kind != NC_FORMAT_NETCDF4) {
-			error("Group list (-g ...) only permitted for netCDF-4 file");
-			exit(EXIT_FAILURE);
-		    }
+		    if(formatting_specs.nc_kind != NC_FORMAT_NETCDF4)
+			goto fail;
 		    /* Check if any grps in -g don't exist */
 		    if(grp_matches(ncid, formatting_specs.nlgrps, formatting_specs.lgrps, formatting_specs.grpids) == 0)
-			exit(EXIT_FAILURE);
+			goto fail;
 		}
 		if (xml_out) {
 		    if(formatting_specs.nc_kind == NC_FORMAT_NETCDF4) {
-			error("NcML output (-x) currently only permitted for netCDF classic model");
-			exit(EXIT_FAILURE);
+			snprintf(errmsg,sizeof(errmsg),"NcML output (-x) currently only permitted for netCDF classic model");
+			goto fail;
 		    }
 		    do_ncdumpx(ncid, path);
 		} else {
@@ -2325,11 +2403,16 @@ main(int argc, char *argv[])
 		}
 	    }
 	    NC_CHECK( nc_close(ncid) );
-	}
-	free(path);
     }
+    nullfree(path) path = NULL;
     exit(EXIT_SUCCESS);
+
+fail: /* ncstat failures */
+    path = (path?path:strdup("<unknown>"));
+    if(ncstat && strlen(errmsg) == 0)
+	snprintf(errmsg,sizeof(errmsg),"%s: %s", path, nc_strerror(ncstat));
+    nullfree(path); path = NULL;
+    if(strlen(errmsg) > 0)
+	error("%s", errmsg);
+    exit(EXIT_FAILURE);
 }
-
-
-END_OF_MAIN()
